/*
 * Authored by Chris Jones <chris.e.jones@gmail.com>
 *
 * Copyright (C) 2006, 2007, 2008 Chris Jones
 *
 * ------------------------------------------------------------------
 *
 * This file is part of the virtkey library.
 *
 * virtkey is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * virtkey is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with virtkey.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef UCS2KEYSYM_H_
#define UCS2KEYSYM_H_

#endif /*UCS2KEYSYM_H_*/
#include <X11/X.h>
#include <stdio.h>

KeySym ucs2keysym(long ucs);
